#include "ble.h"
#include "sd_rpc.h"

#include "ble_hci.h"
#include "ble_central.h"
#include "ble_nus.h"
#include "ble_bts.h"
#include "util.h"

#include <stdbool.h>
#include <stdio.h>
#include <string.h>

// We're not interested in warnings about sprintf() and strcat().
#pragma warning(disable:4996)

// Defined in main.c.
extern adapter_t * m_adapter;
extern char * m_peripheral_mac_address;

state_t m_state = STATE_INITED;

static uint8_t m_connected_devices = 0;

static volatile uint16_t m_connection_handle = 0;

// When discovering characteristics or descriptors, we always need a handle range. The start handle generally increments until it meets the end handle. The end handle is the end handle for the whole service.
static uint16_t m_start_handle_for_service = 0;
static ble_gattc_handle_range_t m_handle_range;

// Nordic UART TX characteristic handle.
static uint16_t m_nus_tx_char_handle = 0;

// Nordic UART RX characteristic handle.
static uint16_t m_nus_rx_char_handle = 0;

// Nordic UART TX characteristic's CCCD handle, which we write to in order to enable notifications.
static uint16_t m_nus_tx_cccd_handle;

// Bike Tracker test state characteristic handle. Write something to this to kick off a test.
static uint16_t m_bts_test_state_char_handle = 0;

// Bike Tracker log characterisitc handle. This is for the full, in memory log, not the realtime log.
static uint16_t m_bts_log_char_handle = 0;

// Bike Tracker log characteristic's CCCD handle. Write something to this to read back the full log all at once.
static uint16_t m_bts_log_cccd_handle = 0;

static volatile bool m_connected = false;
static volatile bool m_connection_is_in_progress = false;
static volatile bool m_write_is_in_progress = false;
static volatile bool m_service_discovery_is_in_progress = false;
static volatile bool m_characteristic_discovery_is_in_progress = false;
static volatile bool m_descriptor_discovery_is_in_progress = false;

static uint16_t m_test_state = 0;
static bool m_test_done = false;

// Uncomment to get verbose printf()s for debugging.
#define DEBUG_FINE

static const ble_gap_scan_params_t m_scan_param =
{
	1, // Active scanning set.
	0, // Selective scanning not set.
	NULL, // White-list not set.
	(uint16_t)SCAN_INTERVAL,
	(uint16_t)SCAN_WINDOW,
	(uint16_t)SCAN_TIMEOUT
};

static const ble_gap_conn_params_t m_connection_param =
{
	(uint16_t)MIN_CONNECTION_INTERVAL,
	(uint16_t)MAX_CONNECTION_INTERVAL,
	(uint16_t)SLAVE_LATENCY,
	(uint16_t)CONNECTION_SUPERVISION_TIMEOUT
};

adapter_t * adapter_init(char * serial_port)
{
	physical_layer_t * phy;
	data_link_layer_t * data_link_layer;
	transport_layer_t * transport_layer;

	phy = sd_rpc_physical_layer_create_uart(serial_port, BAUD_RATE, SD_RPC_FLOW_CONTROL_NONE, SD_RPC_PARITY_NONE);
	data_link_layer = sd_rpc_data_link_layer_create_bt_three_wire(phy, 100);
	transport_layer = sd_rpc_transport_layer_create(data_link_layer, 100);
	return sd_rpc_adapter_create(transport_layer);
}

uint32_t ble_stack_init()
{
	uint32_t err_code;
	ble_enable_params_t ble_enable_params;
	uint32_t * app_ram_base = NULL;

	memset(&ble_enable_params, 0, sizeof(ble_enable_params));

	ble_enable_params.gatts_enable_params.attr_tab_size = BLE_GATTS_ATTR_TAB_SIZE_DEFAULT;
	ble_enable_params.gatts_enable_params.service_changed = false;
	ble_enable_params.gap_enable_params.periph_conn_count = 1;
	ble_enable_params.gap_enable_params.central_conn_count = 1;
	ble_enable_params.gap_enable_params.central_sec_count = 1;
	ble_enable_params.common_enable_params.p_conn_bw_counts = NULL;
	ble_enable_params.common_enable_params.vs_uuid_count = 2;

	err_code = sd_ble_enable(m_adapter, &ble_enable_params, app_ram_base);

	switch (err_code) {
	case NRF_SUCCESS:
		break;
	case NRF_ERROR_INVALID_STATE:
		printf("BLE stack already enabled.\n");
		fflush(stdout);
		break;
	default:
		printf("Failed to enable BLE stack. Error code: %d\n", err_code);
		fflush(stdout);
		break;
	}

	return err_code;
}

uint32_t ble_options_set()
{
	ble_opt_t opt;
	ble_common_opt_t common_opt;

	common_opt.conn_bw.role = BLE_GAP_ROLE_CENTRAL;
	common_opt.conn_bw.conn_bw.conn_bw_rx = BLE_CONN_BW_HIGH;
	common_opt.conn_bw.conn_bw.conn_bw_tx = BLE_CONN_BW_HIGH;
	opt.common_opt = common_opt;

	return sd_ble_opt_set(m_adapter, BLE_COMMON_OPT_CONN_BW, &opt);
}

uint32_t start_scan()
{
	uint32_t error_code = sd_ble_gap_scan_start(m_adapter, &m_scan_param);

	m_state = STATE_SCANNING;

	if (error_code != NRF_SUCCESS)
	{
		printf("Scan start failed.\n");
		fflush(stdout);
	}
	else
	{
		printf("Scan started.\n");
		fflush(stdout);
	}

	return error_code;
}

static void ble_address_to_string_convert(ble_gap_addr_t address, uint8_t * string_buffer)
{
	const int address_length = 6;
	char temp_str[3];

	for (int i = address_length - 1; i >= 0; --i)
	{
		sprintf(temp_str, "%02X", address.addr[i]);
		strcat((char *)string_buffer, temp_str);
	}
}

static void on_adv_report(const ble_gap_evt_t * const p_ble_gap_evt)
{
	uint32_t err_code;
	uint8_t str[STRING_BUFFER_SIZE] = { 0 };

	// Log the MAC address for each advertising packet received.
	ble_address_to_string_convert(p_ble_gap_evt->params.adv_report.peer_addr, str);
	// printf("Adv packet from: 0x%s\n", str);
	fflush(stdout);

	// If this packet is from the peripheral with the MAC address passed in on the command line, connect to it.
	if (strcmp(str, m_peripheral_mac_address) == 0)
	{
		if (m_connected_devices >= MAX_PEER_COUNT || m_connection_is_in_progress)
		{
			return;
		}

		printf("Connecting to 0x%s.\n", str);
		fflush(stdout);

		err_code = sd_ble_gap_connect(m_adapter, &(p_ble_gap_evt->params.adv_report.peer_addr), &m_scan_param, &m_connection_param);

		m_state = STATE_CONNECTING;

		if (err_code != NRF_SUCCESS)
		{
			printf("Error: Connection request failed. Reason: %d\n", err_code);
			fflush(stdout);
			return;
		}

		m_connection_is_in_progress = true;
	}

	// Now expect a call to on_connected(), via ble_evt_dispatch().
}

static uint32_t discover_service(ble_uuid_t * p_srvc_uuid)
{
	uint32_t err_code;
	uint16_t start_handle = 0x01;

	// Discover service. Calls back to on_service_discovery_response() via ble_evt_dispatch().
	m_service_discovery_is_in_progress = true;
	err_code = sd_ble_gattc_primary_services_discover(m_adapter, m_connection_handle, start_handle, p_srvc_uuid);

	if (err_code == BLE_ERROR_INVALID_CONN_HANDLE)
	{
		printf("Invalid connection handle. Are we connected?\n");
		fflush(stdout);
		return err_code;
	}
	else if (err_code == NRF_ERROR_BUSY)
	{
		printf("Error: Busy. Slow down.\n");
		fflush(stdout);
		return err_code;
	}
	else if (err_code != NRF_SUCCESS)
	{
		printf("Failed to start GATT primary service discovery. Error code: 0x%X.\n", err_code);
		fflush(stdout);
		return err_code;
	}

	while (m_service_discovery_is_in_progress)
	{
		// Wait for callback.
	}

	// Do characteristic discovery for this service.
#ifdef DEBUG_FINE
	printf("  Discovering characteristics.\n");
	fflush(stdout);
#endif

	while (m_handle_range.start_handle < m_handle_range.end_handle)
	{
		// Calls back to on_characteristic_discovery_response(), via ble_evt_dispatch().
		m_characteristic_discovery_is_in_progress = true;
		err_code = sd_ble_gattc_characteristics_discover(m_adapter, m_connection_handle, &m_handle_range);

		if (err_code != NRF_SUCCESS)
		{
			printf("Error: sd_ble_gattc_characteristics_discover(): 0x%X.\n", err_code);
			log_error_code(err_code);
			fflush(stdout);

			// Continue
			m_characteristic_discovery_is_in_progress = false;
		}

		while (m_characteristic_discovery_is_in_progress)
		{
			// Wait for callback.
		}

		// Let the callback increment the start handle based on the response.
		// m_handle_range.start_handle++;
	}

	// Do descriptor discovery for the service.
	m_handle_range.start_handle = m_start_handle_for_service;

#ifdef DEBUG_FINE
	printf("  Discovering descriptors.\n");
	fflush(stdout);
#endif

	while (m_handle_range.start_handle < m_handle_range.end_handle)
	{
		// Calls back to on_descriptor_discovery_response(), via ble_evt_dispatch().
		m_descriptor_discovery_is_in_progress = true;
		err_code = sd_ble_gattc_descriptors_discover(m_adapter, m_connection_handle, &m_handle_range);

		if (err_code != NRF_SUCCESS)
		{
			printf("Error: sd_ble_gattc_descriptors_discover(): 0x%X.\n", err_code);
			log_error_code(err_code);
			fflush(stdout);

			// Continue
			m_descriptor_discovery_is_in_progress = false;
		}

		while (m_descriptor_discovery_is_in_progress)
		{
			// Wait for callback.
		}

		// Let the callback increment the start handle based on the response.
		// m_handle_range.start_handle++;
	}

	return err_code;
}

// Do discovery of Nordic's UART service.
uint32_t discover_service_nus()
{
	uint32_t err_code;

	printf("Discovering Nordic UART Service.\n");
	fflush(stdout);

	ble_uuid128_t base_uuid = NUS_BASE_UUID;

	ble_uuid_t srvc_uuid;
	srvc_uuid.type = BLE_UUID_TYPE_VENDOR_BEGIN;
	srvc_uuid.uuid = BLE_UUID_NUS_SERVICE;

	err_code = sd_ble_uuid_vs_add(m_adapter, &base_uuid, &srvc_uuid.type);

	if (err_code != NRF_SUCCESS)
	{
		printf("Failed to add the vendor-specific Nordic UART service. Error code: 0x%X\n.", err_code);
		fflush(stdout);
		return err_code;
	}

	m_state = STATE_DISCOVERING_NUS;

	return discover_service(&srvc_uuid);
}

// Do discovery of the Bike Tracker service.
uint32_t discover_service_bts()
{
	uint32_t err_code;

	printf("Discovering Bike Tracker Service.\n");
	fflush(stdout);

	ble_uuid128_t base_uuid = { BTS_UUID_BASE };

	ble_uuid_t srvc_uuid;
	srvc_uuid.type = BLE_UUID_TYPE_VENDOR_BEGIN;
	srvc_uuid.uuid = BTS_UUID_SERVICE;

	err_code = sd_ble_uuid_vs_add(m_adapter, &base_uuid, &srvc_uuid.type);

	if (err_code != NRF_SUCCESS)
	{
		// NRF_ERROR_NO_MEM (0x04) here means we've init'ed the BLE stack wrong.
		printf("Failed to add the vendor-specific Bike Tracker service. Error code: 0x%X\n.", err_code);
		fflush(stdout);

		return err_code;
	}

	m_state = STATE_DISCOVERING_BTS;

	return discover_service(&srvc_uuid);
}

// Enable/disable notifications on the Nordic UART TX characteristic by writing a 1/0 to its CCCD.
uint32_t nus_tx_cccd_set(uint8_t value)
{
	uint32_t err_code;
	ble_gattc_write_params_t write_params;
	uint8_t cccd_value[2] = { value, 0 };

	if (m_nus_tx_cccd_handle == 0)
	{
		printf("Error: No NUS TX CCCD handle around.\n");
		fflush(stdout);

		return NRF_ERROR_INVALID_STATE;
	}

	write_params.handle = m_nus_tx_cccd_handle;
	write_params.len = 2;
	write_params.p_value = cccd_value;
	write_params.write_op = BLE_GATT_OP_WRITE_REQ;
	write_params.offset = 0;

	printf("Writing to NUS TX CCCD to enable notifications.\n");
	fflush(stdout);

	// Calls back to on_write_response().
	m_write_is_in_progress = true;
	err_code = sd_ble_gattc_write(m_adapter, m_connection_handle, &write_params);

	if (err_code == NRF_ERROR_BUSY)
	{
		printf("Error: Busy. Slow down.\n");
		fflush(stdout);
	}
	else if (err_code != NRF_SUCCESS)
	{
		printf("Failed to write to CCCD. Error code: 0x%X.\n", err_code);
		fflush(stdout);
	}

	return err_code;
}

// Enable/disable notifications on our log characteristic by writing a 1/0 to its CCCD.
uint32_t log_cccd_set(uint8_t value)
{
	uint32_t err_code;
	ble_gattc_write_params_t write_params;
	uint8_t cccd_value[2] = { value, 0 };

#ifdef DEBUG_FINE
	printf("Setting log CCCD.\n");
	fflush(stdout);
#endif

	if (m_bts_log_cccd_handle == 0)
	{
		printf("Error: No log CCCD handle around.\n");
		fflush(stdout);

		return NRF_ERROR_INVALID_STATE;
	}

	write_params.handle = m_bts_log_cccd_handle;
	write_params.len = 2;
	write_params.p_value = cccd_value;
	write_params.write_op = BLE_GATT_OP_WRITE_REQ;
	write_params.offset = 0;

	printf("Writing to log CCCD to enable log notifications.\n");
	fflush(stdout);

	// Calls back to on_write_response().
	m_write_is_in_progress = true;
	err_code = sd_ble_gattc_write(m_adapter, m_connection_handle, &write_params);

	if (err_code == NRF_ERROR_BUSY)
	{
		printf("Error: Busy. Slow down.\n");
		fflush(stdout);
	}
	else if (err_code != NRF_SUCCESS)
	{
		printf("Failed to write to CCCD. Error code: 0x%X.\n", err_code);
		fflush(stdout);
	}

	return err_code;
}

uint32_t start_test(char * p_test_state_name)
{
	// First enable notifications on the NUS TX characteristic.
	// log_cccd_set(BLE_CCCD_NOTIFY);
	nus_tx_cccd_set(BLE_CCCD_NOTIFY);

	while (m_write_is_in_progress)
	{
		// Wait.
	}

	// Note that we don't include the two SIGFOX test states here. They're intended to be run from the mobile app only.
	if (strcmp("pcba", p_test_state_name) == 0 || strcmp("PCBA", p_test_state_name) == 0)
	{
		m_test_state = STATE_TEST_PCBA; // 0x10
	}
	else if (strcmp("system", p_test_state_name) == 0 || strcmp("SYSTEM", p_test_state_name) == 0)
	{
		m_test_state = STATE_TEST_SYSTEM; // 0x11
	}
	else
	{
		printf("Error: No such test state: %s.\n", p_test_state_name);
		return APP_ERROR_NO_SUCH_TEST_STATE;
	}

	ble_gattc_write_params_t write_params;
	write_params.handle = m_bts_test_state_char_handle; // This is decimal 22 at this point.
	write_params.write_op = BLE_GATT_OP_WRITE_REQ;
	write_params.offset = 0;

	// Since the params take a pointer to the value, using the defines directly gets weird. We need the indirection of the variable. The pointer on the write params struct is to a uint8_t, regardless of the length of value we have.
	// TODO: Remove when tested:
	// m_test_state = 0x1;
	write_params.p_value = (uint8_t *)&m_test_state;
	write_params.len = sizeof(uint16_t);
	// write_params.len = 1;

	printf("Writing to test state characteristic to start test %s.\n", p_test_state_name);
	fflush(stdout);

	// Calls back to on_write_response(), via ble_evt_dispatch().
	m_write_is_in_progress = true;
	uint32_t error_code = sd_ble_gattc_write(m_adapter, m_connection_handle, &write_params);

	if (error_code == NRF_ERROR_BUSY)
	{
		printf("Error: Busy. Slow down.\n");
		fflush(stdout);
	}

	return error_code;
}

bool connected()
{
	return m_connected;
}

bool test_done()
{
	return m_test_done;
}

uint32_t close()
{
	uint32_t error_code = sd_rpc_close(m_adapter);

	if (error_code != NRF_SUCCESS)
	{
		printf("Failed to close nRF BLE Driver. Error code: 0x%02X\n", error_code);
		fflush(stdout);
		return error_code;
	}

	printf("Closed.\n");
	fflush(stdout);

	return NRF_SUCCESS;
}

static void on_timeout(const ble_gap_evt_t * const p_ble_gap_evt)
{
	if (p_ble_gap_evt->params.timeout.src == BLE_GAP_TIMEOUT_SRC_CONN)
	{
		m_connection_is_in_progress = false;
	}
	else if (p_ble_gap_evt->params.timeout.src == BLE_GAP_TIMEOUT_SRC_SCAN)
	{
		start_scan();
	}
}

static void on_connected(const ble_gap_evt_t * const p_ble_gap_evt)
{
	printf("Connected.\n");
	fflush(stdout);

	m_connected_devices++;
	m_connection_handle = p_ble_gap_evt->conn_handle;
	m_connection_is_in_progress = false;
	m_connected = true;
}

static void on_disconnected(const ble_gap_evt_t * const p_ble_gap_evt)
{
	// For reason codes, see BLE_HCI_STATUS_CODES in ble_hci.h.
	if (BLE_HCI_CONN_FAILED_TO_BE_ESTABLISHED)
	{
		printf("Disconnected. Connection failed to be established. Is the UUT already bonded with a central?\n");
		fflush(stdout);
	}
	else
	{
		printf("Disconnected. Reason: 0x%02X\n", p_ble_gap_evt->params.disconnected.reason);
		fflush(stdout);
	}

	m_connected_devices--;
	m_connection_handle = 0;
	m_connection_is_in_progress = false;
	m_connected = false;
}

static void on_service_discovery_response(const ble_gattc_evt_t * const p_ble_gattc_evt)
{
#ifdef DEBUG_FINE
	printf("Got service discovery response.\n");
	fflush(stdout);
#endif

	int count;
	int service_index;
	const ble_gattc_service_t * service;

	if (p_ble_gattc_evt->gatt_status != NRF_SUCCESS)
	{
		printf("Service discovery failed. Error code 0x%X\n", p_ble_gattc_evt->gatt_status);
		fflush(stdout);
		return;
	}

	count = p_ble_gattc_evt->params.prim_srvc_disc_rsp.count;

	if (count == 0)
	{
		printf("Service not found\n");
		fflush(stdout);
		return;
	}

	// We expect to discover only the service we requested.
	if (count > 1)
	{
		printf("Warning: Discovered multiple primary services. Ignoring all but the first.\n");
		fflush(stdout);
	}

	service_index = 0;
	service = &(p_ble_gattc_evt->params.prim_srvc_disc_rsp.services[service_index]);

	m_start_handle_for_service = service->handle_range.start_handle;

	m_handle_range.start_handle = service->handle_range.start_handle;
	m_handle_range.end_handle = service->handle_range.end_handle;

#ifdef DEBUG_FINE
	printf("Discovered service. UUID: 0x%04X, start handle: 0x%04X, end handle: 0x%04X\n", service->uuid.uuid, m_handle_range.start_handle, m_handle_range.end_handle);
	fflush(stdout);
#endif

	m_service_discovery_is_in_progress = false;
}

static void on_characteristic_discovery_response(const ble_gattc_evt_t * const p_ble_gattc_evt)
{
	int count = p_ble_gattc_evt->params.char_disc_rsp.count;

	if (p_ble_gattc_evt->gatt_status != NRF_SUCCESS)
	{
		printf("Error: Characteristic discovery failed. Error: 0x%X.\n", p_ble_gattc_evt->gatt_status);
		fflush(stdout);
		return;
	}

	ble_gattc_char_t c;

	for (int i = 0; i < count; i++)
	{
		c = p_ble_gattc_evt->params.char_disc_rsp.chars[i];

#ifdef DEBUG_FINE
		printf("  Characteristic handle: 0x%04X, UUID: 0x%04X (%d of %d).\n", c.handle_decl, c.uuid.uuid, i + 1, count);
		fflush(stdout);
#endif

		if (c.uuid.uuid == BLE_UUID_NUS_TX_CHARACTERISTIC)
		{
#ifdef DEBUG_FINE
			printf("  NUS TX.\n");
			fflush(stdout);
#endif

			m_nus_tx_char_handle = c.handle_decl;
		}
		else if (c.uuid.uuid == BLE_UUID_NUS_RX_CHARACTERISTIC)
		{
#ifdef DEBUG_FINE
			printf("  NUS RX.\n");
			fflush(stdout);
#endif

			m_nus_rx_char_handle = c.handle_decl;
		}
		else if (c.uuid.uuid == BTS_UUID_TEST_STATE_CHAR)
		{
#ifdef DEBUG_FINE
			printf("  BTS test state.\n");
			fflush(stdout);
#endif

			m_bts_test_state_char_handle = c.handle_decl;
		}
		else if (c.uuid.uuid == BTS_UUID_LOG_CHAR)
		{
#ifdef DEBUG_FINE
			printf("  BTS log.\n");
			fflush(stdout);
#endif

			m_bts_log_char_handle = c.handle_decl;
		}

		// Increment the start handle for the next discovery call.
		m_handle_range.start_handle = c.handle_value + 1;
	}

	m_characteristic_discovery_is_in_progress = false;
}

static void on_descriptor_discovery_response(const ble_gattc_evt_t * const p_ble_gattc_evt)
{
	int count = p_ble_gattc_evt->params.desc_disc_rsp.count;

	if (p_ble_gattc_evt->gatt_status != NRF_SUCCESS)
	{
		printf("Error: Descriptor discovery failed. Error code: 0x%X\n", p_ble_gattc_evt->gatt_status);
		fflush(stdout);
		return;
	}

	for (int i = 0; i < count; i++)
	{
		ble_gattc_desc_t d = p_ble_gattc_evt->params.desc_disc_rsp.descs[i];

#ifdef DEBUG_FINE
		printf("  Descriptor handle: 0x%04X, UUID: 0x%04X (%d of %d)\n", d.handle, d.uuid.uuid, i + 1, count);
		fflush(stdout);
#endif

		if (d.uuid.uuid == BLE_UUID_CCCD)
		{
			if (m_state == STATE_DISCOVERING_NUS)
			{
#ifdef DEBUG_FINE
				printf("  CCCD (nus).\n");
				fflush(stdout);
#endif

				m_nus_tx_cccd_handle = d.handle;
			}
			else if (m_state == STATE_DISCOVERING_BTS)
			{
#ifdef DEBUG_FINE
				printf("  CCCD (bts).\n");
				fflush(stdout);
#endif

				m_bts_log_cccd_handle = d.handle;
			}
		}

		// Increment the start handle for the next discovery call.
		m_handle_range.start_handle = d.handle + 1;

		m_descriptor_discovery_is_in_progress = false;
	}
}

static void on_write_response(const ble_gattc_evt_t * const p_ble_gattc_evt)
{
	if (p_ble_gattc_evt->gatt_status == NRF_SUCCESS)
	{
#ifdef DEBUG_FINE
		printf("Write completed OK.\n");
		fflush(stdout);
#endif
	}
	else if (p_ble_gattc_evt->gatt_status == BLE_GATT_STATUS_ATTERR_WRITE_NOT_PERMITTED)
	{
		printf("Error: Write not permitted. Is this the right characteristic or descriptor?\n");
		fflush(stdout);
	}
	else
	{
		printf("Error: Write operation failed. Error code 0x%X\n", p_ble_gattc_evt->gatt_status);
		fflush(stdout);
	}

	m_write_is_in_progress = false;
}

// hvx: handle Value x, where x is indication or notification.
static void on_hvx(const ble_gattc_evt_t * const p_ble_gattc_evt)
{
	// Bike Tracker log characteristic and CCCD.
	if (p_ble_gattc_evt->params.hvx.handle >= m_bts_log_char_handle || p_ble_gattc_evt->params.hvx.handle <= m_bts_log_cccd_handle)
	{
		char * line = "";
		memcpy(line, p_ble_gattc_evt->params.hvx.data, p_ble_gattc_evt->params.hvx.len);

		// Our sole purpose in life is to emit this log for the calling process to read back and parse for errors.
		printf("%s", line);

		// Look out for END_OF_TEST. The whole process should complete once we've seen it.
		if (strstr(line, "END_OF_TEST") != NULL)
		{
			m_test_done = true;
		}
	}
	else
	{
		printf("Un-parsed data received on handle: %04X\n", p_ble_gattc_evt->params.hvx.handle);
	}

	fflush(stdout);
}

static void on_conn_params_update_request(const ble_gap_evt_t * const p_ble_gap_evt)
{
	uint32_t err_code = sd_ble_gap_conn_param_update(m_adapter, m_connection_handle, &(p_ble_gap_evt->params.conn_param_update_request.conn_params));

	if (err_code != NRF_SUCCESS)
	{
		// Error 0x3002 here is this, from ble_err.h:
		// #define BLE_ERROR_INVALID_CONN_HANDLE    (NRF_ERROR_STK_BASE_NUM+0x002) /**< Invalid connection handle. */.
		printf("Warning: Conn params update failed. Error: %d\n", err_code);
		fflush(stdout);
	}
}

static uint32_t adv_report_parse(uint8_t type, data_t * p_advdata, data_t * p_typedata)
{
	uint32_t  index = 0;
	uint8_t * p_data;

	p_data = p_advdata->p_data;

	while (index < p_advdata->data_len)
	{
		uint8_t field_length = p_data[index];
		uint8_t field_type = p_data[index + 1];

		if (field_type == type)
		{
			p_typedata->p_data = &p_data[index + 2];
			p_typedata->data_len = field_length - 1;
			return NRF_SUCCESS;
		}

		index += field_length + 1;
	}

	return NRF_ERROR_NOT_FOUND;
}

void ble_evt_dispatch(adapter_t * adapter, ble_evt_t * p_ble_evt)
{
	if (p_ble_evt == NULL)
	{
		printf("Error: Empty BLE event.\n");
		fflush(stdout);
		return;
	}

	switch (p_ble_evt->header.evt_id)
	{
	case BLE_GAP_EVT_CONNECTED:
		on_connected(&(p_ble_evt->evt.gap_evt));
		break;

	case BLE_GAP_EVT_DISCONNECTED:
		on_disconnected(&(p_ble_evt->evt.gap_evt));
		break;

	case BLE_GAP_EVT_ADV_REPORT:
		on_adv_report(&(p_ble_evt->evt.gap_evt));
		break;

	case BLE_GAP_EVT_TIMEOUT:
		on_timeout(&(p_ble_evt->evt.gap_evt));
		break;

	case BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP:
		on_service_discovery_response(&(p_ble_evt->evt.gattc_evt));
		break;

	case BLE_GATTC_EVT_CHAR_DISC_RSP:
		on_characteristic_discovery_response(&(p_ble_evt->evt.gattc_evt));
		break;

	case BLE_GATTC_EVT_DESC_DISC_RSP:
		on_descriptor_discovery_response(&(p_ble_evt->evt.gattc_evt));
		break;

	case BLE_GATTC_EVT_WRITE_RSP:
		on_write_response(&(p_ble_evt->evt.gattc_evt));
		break;

	case BLE_GATTC_EVT_HVX:
		on_hvx(&(p_ble_evt->evt.gattc_evt));
		break;

	case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST:
		on_conn_params_update_request(&(p_ble_evt->evt.gap_evt));
		break;

	case BLE_GAP_EVT_CONN_PARAM_UPDATE: // Connection Parameters updated.
	case BLE_GAP_EVT_SEC_PARAMS_REQUEST: // Request to provide security parameters
	case BLE_GAP_EVT_SEC_INFO_REQUEST: // Request to provide security information.
	case BLE_GAP_EVT_PASSKEY_DISPLAY: // Request to display a passkey to the user.
	case BLE_GAP_EVT_KEY_PRESSED: // Notification of a keypress on the remote device.
	case BLE_GAP_EVT_AUTH_KEY_REQUEST: // Request to provide an authentication key.
	case BLE_GAP_EVT_LESC_DHKEY_REQUEST: // Request to calculate an LE Secure Connections DHKey.
	case BLE_GAP_EVT_AUTH_STATUS: // Authentication procedure completed with status
	case BLE_GAP_EVT_CONN_SEC_UPDATE: // Connection security updated.
	case BLE_GAP_EVT_RSSI_CHANGED: // RSSI report.
	case BLE_GAP_EVT_SEC_REQUEST: // Security Request.
	case BLE_GAP_EVT_SCAN_REQ_REPORT: // Scan request report.
		// Intentionally unhandled.
		break;

	default:
		printf("Warning: Unhandled BLE event with ID: %d\n", p_ble_evt->header.evt_id);
		fflush(stdout);
		break;
	}
}
